class Config:
	NOMECLATURA = ["I1", "I2", "I3", "I4", "I5", "I6", "I7", "I8", "I9", "I10"]
	PRIORIDADE = [0.1526, 0.1121, 0.0997, 0.2269, 0.0992, 0.0831, 0.0954, 0.0756, 0.0277, 0.0277]
	INDICADOR = [
		"indicador_via_pavimentada",
		"distancia",
		"tipo_integracao",
		"pontualidade",
		"frequencia_atendimento_pontuacao",
		"cumprimento_itinerario",
		"proporcao",
		"indicador_treinamento_motorista",
		"disponibilidade_informacao",
		"valor_tarifa",
	]

config = Config()
